from pathlib import Path
from datetime import datetime, date, time
from typing import Literal, TypeVar


FileType = Literal[
    "text",
    "image",
    "video",
    "audio",
    "document",
]

FileFormat = Literal[
    "json",
    "jsonl",
    "toml",
    "yaml",
]

StandardFileExtension = FileFormat


VariantFileExtension = Literal[
    "ndjson",
    "jsonlines",
    "yml",
]

FileExtension = StandardFileExtension | VariantFileExtension

MimeType = Literal[
    "application/json",
    "application/jsonl",
]

FilePath = str | Path

# DataLocation = Literal[
#     "disk",
#     "web",
#     "memory",
# ]


# ! Python Data Types

JSON_TYPE = TypeVar("JSON_TYPE", bound="PyJSON")
PyJSON = bool | int | float | str | list[JSON_TYPE] | dict[str, JSON_TYPE]
SerializedJSON = str | bytes | bytearray

TOML_TYPE = TypeVar("TOML_TYPE", bound="TOML_VALUE_TYPE")
TOML_VALUE_TYPE = (
    bool
    | int
    | float
    | str
    | datetime
    | date
    | time
    | list[TOML_TYPE]
    | dict[str, TOML_TYPE]
)
PyTOML = dict[str, TOML_VALUE_TYPE]
SerializedTOML = str | bytes

YAML_TYPE = TypeVar("YAML_TYPE", bound="PyYAML")
PyYAML = bool | int | float | str | list[YAML_TYPE] | dict[str, YAML_TYPE]
SerializedYAML = str | bytes

# ! ---



from pathlib import Path
from pyeio.common.types import FilePath, PyTOML, SerializedTOML
from pyeio.common.exceptions import MissingExtraError
from pyeio.file import io, utils
from pyeio.io import web

try:
    import toml
except ImportError:
    raise MissingExtraError(extra="toml")


def parse(data: SerializedTOML) -> PyTOML:
    if isinstance(data, bytes):
        # todo.fix: will not always work
        data = str(data.decode())
    if not isinstance(data, str):
        # todo: add details on type error
        raise TypeError()
    toml_data = toml.loads(data)
    return toml_data


def serialize(data: PyTOML) -> str:
    serialized_data = toml.dumps(data)
    return serialized_data


def open(
    path: FilePath,
    validate_file_extension: bool = True,
) -> PyTOML:
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="toml",
            path=path,
        )
    serialized_toml = io.open_text(path=path)
    return parse(serialized_toml)


def save(
    data: PyTOML,
    path: FilePath,
    overwrite: bool = False,
    validate_file_extension: bool = True,
) -> None:
    """
    Save data to a TOML file.

    Args:
        data (PyTOML): TOML-compatible Python data structure to serialize and save.
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
    """
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="toml",
            path=path,
        )
    serialized_toml: str = serialize(data)
    io.save_text(data=serialized_toml, path=path, overwrite=overwrite)


def load(
    url: str,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool | None = None,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> PyTOML:
    """
    Load a TOML file from a URL directly into memory.

    Args:
        url (str): The URL of the TOML file.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
        evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
        show_progress (bool, optional): Show progress bar. Defaults to False.
        show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

    Returns:
        PyTOML: TOML data as built-in Python data structures.
    """
    if evaluate_size is None:
        evaluate_size = show_progress
    binary_data = web.load_binary(
        url=url,
        chunk_size=chunk_size,
        follow_redirects=follow_redirects,
        evaluate_size=evaluate_size,
        show_progress=show_progress,
        show_file_name=show_file_name,
    )
    serialized_data = binary_data.decode("utf-8")
    toml_data = parse(serialized_data)
    return toml_data


def download(
    url: str,
    path: FilePath,
    overwrite: bool = False,
    validate_file_extension: bool = True,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool | None = None,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> None:
    """
    Download a TOML file from a URL directly to a local TOML file on disk.

    Args:
        url (str): The URL of the TOML file.
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
        evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
        show_progress (bool, optional): Show progress bar. Defaults to False.
        show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

    Raises:
        FileExistsError: File already exists. Set `overwrite` to `True` to bypass.
    """
    if evaluate_size is None:
        evaluate_size = show_progress
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="toml",
            path=path,
        )
    if path.exists() and (not overwrite):
        raise FileExistsError(str(path))
    toml_data = load(
        url=url,
        chunk_size=chunk_size,
        follow_redirects=follow_redirects,
        evaluate_size=evaluate_size,
        show_progress=show_progress,
        show_file_name=show_file_name,
    )
    save(
        data=toml_data,
        path=path,
        overwrite=overwrite,
        validate_file_extension=validate_file_extension,
    )



from typing import Optional
from pathlib import Path
from urllib.parse import urlparse
from pyeio.common.types import FilePath
from pyeio import __meta__


class UnexpectedError(Exception):
    """Raised when something unexpected happens."""

    def __init__(
        self,
        details: Optional[str] = None,
        issues: bool = True,
    ) -> None:
        self.message = "This was unexpected!"
        if details:
            self.message += f"\nDetails: {details}"
        if issues:
            self.message += f"\nPlease submit a GitHub issue with the code that generated this error at:\n{__meta__.__issues__}"
        super().__init__(self.message)


class IncorrectFileExtensionError(Exception):
    """Raised when the provided extension does not match the expected extension(s)."""

    def __init__(
        self,
        file_extension: str,
        compatible_extensions: set[str],
    ) -> None:
        self.file_extension = file_extension
        self.compatible_extensions = compatible_extensions
        self.message = f"Extension '{self.file_extension}' should be in '{self.compatible_extensions}'"
        super().__init__(self.message)


class MissingExtraError(Exception):
    def __init__(self, extra: str, *args: object) -> None:
        self.message = f"To use this module install: '{__meta__.__package__}[{extra}]'"
        super().__init__(*args)


class MissingFileExtensionError(Exception):
    """Raised when the file in question doesn't have an extension."""

    def __init__(
        self,
        file: str,
    ) -> None:
        self.message = f"'{file}' does not have a file extension. You may be able to bypass this by passing 'False' to the parameter that validates file extensions."
        super().__init__(self.message)


class UnsupportedFileExtensionError(Exception):
    """Raised when the provided extension is not known/supported yet."""

    def __init__(
        self,
        extension: str,
    ) -> None:
        self.extension = extension
        self.message = f"'{self.extension}' is not yet supported."
        super().__init__(self.message)



from pathlib import Path
from typing import Callable, Generator, Any
from zstandard import ZstdDecompressor


MAX_WINDOW_SIZE: int = 1 << 31


class Streamer: ...


class Reader:
    def __init__(
        self,
        file_path: str | Path,
        chunk_delimiter: bytes,
        chunk_parser: Callable[[bytes], Any] = lambda x: x,
        block_size_bytes: int = 1 << 20,
    ) -> None:
        self.file_path = file_path
        self.chunk_delimiter = chunk_delimiter
        self.chunk_parser = chunk_parser
        self.block_size_bytes = block_size_bytes
        self.reset()

    def reset(self) -> None:
        self.stream = ZstdDecompressor(
            max_window_size=MAX_WINDOW_SIZE,
        ).stream_reader(open(self.file_path, "rb"))
        self.buffer = b""
        self.chunks = []

    def __iter__(self) -> Generator[bytes, None, None]:
        while True:
            try:
                yield next(self)
            except StopIteration:
                break

    def __next__(self) -> bytes | Any:
        if len(self.chunks):
            current = self.chunks.pop(0)
            return self.chunk_parser(current)
        else:
            chunk = self.stream.read(self.block_size_bytes)
            if chunk:
                self.chunks = (self.buffer + chunk).split(self.chunk_delimiter)
                self.buffer = self.chunks[-1]
                self.chunks = self.chunks[:-1]
                current = self.chunks.pop(0)
                return self.chunk_parser(current)
            else:
                raise StopIteration()

    def read_chunk(self) -> bytes | Any:
        return self.__next__()

    def read_chunks(self, n: int) -> list[bytes] | list[Any]:
        return [self.read_chunk() for _ in range(n)]


def read(
    file_path: str | Path,
    chunk_delimiter: bytes = b"\n",
    block_size: int = 1 << 20,
    parser: Callable[[bytes], Any] = lambda x: x,
) -> Generator[bytes, None, None] | Generator[Any, None, None]:
    reader = Reader(
        file_path=file_path,
        chunk_delimiter=chunk_delimiter,
        block_size_bytes=block_size,
        chunk_parser=parser,
    )
    for chunk in reader:
        yield chunk


def compress():
    """Compress an existing file to ZST."""
    raise NotImplementedError()


def decompress():
    """Decompress an existing file."""
    raise NotImplementedError()


# def load():
#     """Decompress and load entire file into memory as bytes."""
#     raise NotImplementedError()


def save(data: bytes, path: str | Path) -> None:
    """Compress and save serializable data to .zst file."""
    raise NotImplementedError()


# from pyeio.core.types import FilePath


# def read(): ...


# def read_lines(): ...


# def compress(source: FilePath, target: FilePath): ...


# def decompress(source: FilePath, target: FilePath): ...



import yaml
from pathlib import Path
from pyeio.common.types import FilePath, PyYAML, SerializedYAML
from pyeio.common.exceptions import MissingExtraError
from pyeio.file import io, utils
from pyeio.io import web


try:
    import yaml
except ImportError:
    raise MissingExtraError(extra="yaml")


# todo: add and account for yaml base lib parameters


def parse(data: SerializedYAML) -> PyYAML:
    if isinstance(data, bytes):
        # todo: add chardet detection
        data = data.decode("utf-8")
    if not isinstance(data, str):
        raise TypeError(
            f"Expected data to be of type 'str' or 'bytes', got {type(data)}"
        )
    yaml_data = yaml.safe_load(data)
    return yaml_data


def serialize(data: PyYAML, default_flow_style: bool = False) -> str:
    serialized_data = yaml.dump(data, default_flow_style=default_flow_style)
    return serialized_data


def open(
    path: FilePath,
    validate_file_extension: bool = True,
) -> PyYAML:
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="yaml",
            path=path,
        )
    serialized_yaml = io.open_text(path=path)
    return parse(serialized_yaml)


def save(
    data: PyYAML,
    path: FilePath,
    overwrite: bool = False,
    default_flow_style: bool = False,
    validate_file_extension: bool = True,
) -> None:
    """
    Save data to a YAML file.

    Args:
        data (PyYAML): YAML-compatible Python data structure to serialize and save.
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
        default_flow_style (bool, optional): Set to True to use the block style for nested collections. Defaults to False.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
    """
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="yaml",
            path=path,
        )
    serialized_yaml: str = serialize(data, default_flow_style=default_flow_style)
    io.save_text(data=serialized_yaml, path=path, overwrite=overwrite)


def load(
    url: str,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool | None = None,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> PyYAML:
    """
    Load a YAML file from a URL directly into memory.

    Args:
        url (str): The URL of the YAML file.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
        evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
        show_progress (bool, optional): Show progress bar. Defaults to False.
        show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

    Returns:
        PyYAML: YAML data as built-in Python data structures.
    """
    if evaluate_size is None:
        evaluate_size = show_progress
    binary_data = web.load_binary(
        url=url,
        chunk_size=chunk_size,
        follow_redirects=follow_redirects,
        evaluate_size=evaluate_size,
        show_progress=show_progress,
        show_file_name=show_file_name,
    )
    serialized_data = binary_data.decode("utf-8")
    yaml_data = parse(serialized_data)
    return yaml_data


def download(
    url: str,
    path: FilePath,
    overwrite: bool = False,
    default_flow_style: bool = False,
    validate_file_extension: bool = True,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool | None = None,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> None:
    """
    Download a YAML file from a URL directly to a local YAML file on disk.

    Args:
        url (str): The URL of the YAML file.
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
        default_flow_style (bool, optional): Set to True to use the block style for nested collections. Defaults to False.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
        evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
        show_progress (bool, optional): Show progress bar. Defaults to False.
        show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

    Raises:
        FileExistsError: File already exists. Set `overwrite` to `True` to bypass.
    """
    if evaluate_size is None:
        evaluate_size = show_progress
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="yaml",
            path=path,
        )
    if path.exists() and not overwrite:
        raise FileExistsError(str(path))
    yaml_data = load(
        url=url,
        chunk_size=chunk_size,
        follow_redirects=follow_redirects,
        evaluate_size=evaluate_size,
        show_progress=show_progress,
        show_file_name=show_file_name,
    )
    save(
        data=yaml_data,
        path=path,
        overwrite=overwrite,
        default_flow_style=default_flow_style,
        validate_file_extension=validate_file_extension,
    )



from typing import Generator
from pyeio.common.types import FilePath


class ChunkReader:
    """For reading chunks of data with defined number of bytes in each chunk."""

    def __init__(self, path: FilePath, size: int = 1024) -> None:
        self.path = path
        self.size = size
        self.file = open(self.path, "rb")

    def __next__(self) -> bytes:
        chunk = self.file.read(self.size)
        if chunk:
            return chunk
        else:
            raise StopIteration()

    def __iter__(self) -> Generator[bytes, None, None]:
        while True:
            try:
                yield next(self)
            except StopIteration:
                break

    def reset(self) -> None:
        self.file.seek(0)


class BlockReader:
    """Reads blocks based on whether they match the delimiter pattern"""

    ...


class ParseReader: ...


class TokenReader: ...



import json
from pathlib import Path
from pyeio.common.types import FilePath, PyJSON, SerializedJSON
from pyeio.file import utils
from pyeio.io import web, loc


def parse(data: SerializedJSON) -> PyJSON:
    """
    Just calls `json.loads`. Included to increase API standardization between different formats.

    Args:
        data (SerializedJSON): Serialized JSON data as type (`str`, `bytes`, or `bytearray`).

    Returns:
        PyJSON: JSON data as built-in python data structure(s).
    """
    json_data = json.loads(data)
    return json_data


def serialize(
    data: PyJSON,
    skipkeys: bool = False,
    ensure_ascii: bool = False,
    check_circular: bool = True,
    allow_nan: bool = True,
    sort_keys: bool = True,
    compact_serialization: bool = True,
) -> str:
    """
    Just calls `json.dumps`. Included to increase API standardization between different formats.

    Args:
        data (PyJSON): JSON data to serialize.
        skipkeys (bool, optional): Skip `dict` keys that are not in (`str`,`int`,`float`,`bool`,`None`) instead of raising `TypeError`. Defaults to False.
        ensure_ascii (bool, optional): Ensure the serialized string only contains ascii characters. Defaults to False.
        check_circular (bool, optional): Check circular reference for container types to avoid `RecursionError`. Defaults to True.
        allow_nan (bool, optional): Allow (`nan`, `inf`, `-inf`) values. Defaults to True.
        sort_keys (bool, optional): Sort the output dictionaries by key. Defaults to True.
        compact_serialization (bool, optional): Make the serialization compact. Set to `False` to pretty format. Defaults to True.

    Returns:
        str: Serialized JSON data.
    """
    indent = None if compact_serialization else 4
    separators = (",", ":") if compact_serialization else (", ", ": ")
    serialized_data = json.dumps(
        data,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        indent=indent,
        separators=separators,
        sort_keys=sort_keys,
    )
    return serialized_data


def open(
    path: FilePath,
    validate_file_extension: bool = True,
) -> PyJSON:
    """
    Open a JSON file.

    Args:
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.

    Returns:
        PyJSON: JSON data as built-in python data structure(s).
    """
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="json",
            path=path,
        )
    serialized_json: str = loc.open_text(path=path)
    json_data = parse(serialized_json)
    return json_data


def save(
    data: PyJSON,
    path: FilePath,
    overwrite: bool = False,
    skipkeys: bool = False,
    ensure_ascii: bool = False,
    check_circular: bool = True,
    allow_nan: bool = True,
    sort_keys: bool = True,
    compact_serialization: bool = True,
    validate_file_extension: bool = True,
) -> None:
    """
    Save data to a JSON file.

    Args:
        data (PyJSON): JSON compatible python data structure data to serialize and save.
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
        data (PyJSON): JSON data to serialize.
        skipkeys (bool, optional): Skip `dict` keys that are not in (`str`,`int`,`float`,`bool`,`None`) instead of raising `TypeError`. Defaults to False.
        ensure_ascii (bool, optional): Ensure the serialized string only contains ascii characters. Defaults to False.
        check_circular (bool, optional): Check circular reference for container types to avoid `RecursionError`. Defaults to True.
        allow_nan (bool, optional): Allow (`nan`, `inf`, `-inf`) values. Defaults to True.
        sort_keys (bool, optional): Sort the output dictionaries by key. Defaults to True.
        compact_serialization (bool, optional): Make the serialization compact. Set to `False` to pretty format. Defaults to True.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
    """
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="json",
            path=path,
        )
    serialized_json: str = serialize(
        data=data,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        sort_keys=sort_keys,
        compact_serialization=compact_serialization,
    )
    loc.save_text(data=serialized_json, path=path, overwrite=overwrite)


def load(
    url: str,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool | None = None,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> PyJSON:
    """
    Load a JSON file at a URL directly into memory.

    Args:
        url (str): The URL of the JSON file.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
        evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
        show_progress (bool, optional): Show progress bar. Defaults to False.
        show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

    Returns:
        PyJSON: JSON data as built-in python data structure(s).
    """
    if evaluate_size is None:
        evaluate_size = show_progress
    binary_data = web.load_binary(
        url=url,
        chunk_size=chunk_size,
        follow_redirects=follow_redirects,
        evaluate_size=evaluate_size,
        show_progress=show_progress,
        show_file_name=show_file_name,
    )
    # todo.fix: need to do character detection and account for when utf-8 fails
    serialized_data = binary_data.decode("utf-8")
    # todo.ext: additionally, do mime detection, data detection, add a bunch of params for exp features
    json_data = parse(serialized_data)
    return json_data


def download(
    url: str,
    path: FilePath,
    overwrite: bool = False,
    skipkeys: bool = False,
    ensure_ascii: bool = False,
    check_circular: bool = True,
    allow_nan: bool = True,
    sort_keys: bool = True,
    compact_serialization: bool = True,
    validate_file_extension: bool = True,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool | None = None,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> None:
    """
    Download a JSON file at a URL directly to a local JSON file on disk.

    Args:
        url (str): The URL of the JSON file.
        path (FilePath): The path to the file as a `str` or `pathlib.Path`.
        overwrite (bool, optional): Allow existing file to be overwritten. Defaults to False.
        data (PyJSON): JSON data to serialize.
        skipkeys (bool, optional): Skip `dict` keys that are not in (`str`,`int`,`float`,`bool`,`None`) instead of raising `TypeError`. Defaults to False.
        ensure_ascii (bool, optional): Ensure the serialized string only contains ascii characters. Defaults to False.
        check_circular (bool, optional): Check circular reference for container types to avoid `RecursionError`. Defaults to True.
        allow_nan (bool, optional): Allow (`nan`, `inf`, `-inf`) values. Defaults to True.
        sort_keys (bool, optional): Sort the output dictionaries by key. Defaults to True.
        compact_serialization (bool, optional): Make the serialization compact. Set to `False` to pretty format. Defaults to True.
        validate_file_extension (bool, optional): Validate the file extension is supported and correct. Defaults to True.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.
        evaluate_size (bool, optional): Evaluate the size of the resource before downloading. Defaults to None.
        show_progress (bool, optional): Show progress bar. Defaults to False.
        show_file_name (bool, optional): Show name of file in progress bar if used. Defaults to True.

    Raises:
        FileExistsError: File already exists. Set `overwrite` to `True` to bypass.
    """
    if evaluate_size is None:
        evaluate_size = show_progress
    path = Path(path)
    if validate_file_extension:
        utils.run_file_extension_validation(
            module_name="json",
            path=path,
        )
    if path.exists() and (not overwrite):
        raise FileExistsError(str(path))
    json_data = load(
        url=url,
        chunk_size=chunk_size,
        follow_redirects=follow_redirects,
        evaluate_size=evaluate_size,
        show_progress=show_progress,
        show_file_name=show_file_name,
    )
    save(
        data=json_data,
        path=path,
        overwrite=overwrite,
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        sort_keys=sort_keys,
        compact_serialization=compact_serialization,
        validate_file_extension=validate_file_extension,
    )



import httpx
from io import BytesIO
from pathlib import Path
from typing import Optional
from tqdm import tqdm
from pyeio.file import utils

# todo: add user agent as parameter to remaining functions
# todo: add cookies to functions
# todo.ext: httpx params


def _add_custom_user_agent_to_headers(
    user_agent: Optional[str],
    headers: dict[str, str],
) -> dict[str, str]:
    if not (user_agent is None):
        headers["user-agent"] = user_agent
    return headers


def check_if_remote_accepts_byte_range(
    url: str,
    follow_redirects: bool = True,
    user_agent: Optional[str] = None,
) -> bool:
    """
    Check if a remote server accepts byte range requests.

    This function sends a HEAD request to the specified URL with a 'Range' header
    to determine if the server supports partial content requests.

    Args:
        url (str): The URL to check for byte range support.
        follow_redirects (bool, optional): Whether to follow redirects. Defaults to True.
        user_agent (str, optional): Optional custom user agent.

    Returns:
        bool: True if the server accepts byte range requests, False otherwise.
    """
    headers = {"Range": "bytes=0-8"}
    headers = _add_custom_user_agent_to_headers(user_agent, headers)
    response = httpx.head(url=url, headers=headers, follow_redirects=follow_redirects)
    response.raise_for_status()
    return response.status_code == 206


def load_binary_chunk(
    url: str,
    start: int,
    end: int,
    follow_redirects: bool = True,
    user_agent: Optional[str] = None,
) -> bytes:
    """
    Read a specific byte range from a remote resource.

    This function sends a GET request to the specified URL with a 'Range'
    header to retrieve a specific chunk of the resource.

    Args:
        url (str): The URL of the remote resource.
        start (int): The starting byte position of the chunk.
        end (int): The ending byte position of the chunk.
        follow_redirects (bool, optional): Whether to follow redirects. Defaults to True.
        user_agent (str, optional): Optional custom user agent.

    Returns:
        bytes: The requested chunk of the resource as bytes.

    Raises:
        Exception: If the remote server does not accept byte range requests.
    """
    headers = {"Range": f"bytes={start}-{end}"}
    headers = _add_custom_user_agent_to_headers(user_agent, headers)
    buffer = BytesIO()
    response = httpx.get(url, headers=headers, follow_redirects=follow_redirects)
    response.raise_for_status()
    if response.status_code != 206:
        raise Exception("The remote server does not accept byte range requests.")
    buffer.write(response.content)
    buffer.seek(0)
    return buffer.read()


def request_content_length(
    url: str,
    follow_redirects: bool = True,
    user_agent: Optional[str] = None,
) -> int:
    """
    Get the content length of a remote resource.

    This function sends a HEAD request to the specified URL and retrieves
    the 'Content-Length' header value.

    Args:
        url (str): The URL of the remote resource.
        follow_redirects (bool, optional): Whether to follow redirects. Defaults to True.
        user_agent (str, optional): Optional custom user agent.

    Returns:
        int: The content length of the resource in bytes.

    Raises:
        KeyError: If the 'Content-Length' header is not found in the response.
    """
    headers = dict()
    headers = _add_custom_user_agent_to_headers(user_agent, headers)
    response = httpx.head(url, headers=headers, follow_redirects=follow_redirects)
    response.raise_for_status()
    headers = {k.lower(): v for k, v in dict(response.headers).items()}
    if "content-length" not in headers.keys():
        raise KeyError("'content-length' header not found")
    else:
        return int(headers["content-length"])


def estimate_resource_size(
    url: str,
    initial_guess: int = 1 << 20,
    max_requests: int = 10,
    user_agent: Optional[str] = None,
) -> int:
    """
    Estimates the content length of a resource by making strategic byte range requests.

    Args:
        url (str): The URL of the resource.
        initial_guess (int): Initial guess for the file size in bytes. Default is 1MB.
        max_requests (int): Maximum number of requests to make before giving up.
        user_agent (str, optional): Optional custom user agent.

    Returns:
        int: Estimated content length in bytes
    """
    headers = dict()
    headers = _add_custom_user_agent_to_headers(user_agent, headers)
    with httpx.Client(headers=headers) as httpx_client:
        lower_bound = 0
        upper_bound = initial_guess
        requests_made = 0
        while requests_made < max_requests:
            requests_made += 1
            mid = (lower_bound + upper_bound) // 2
            headers = {"Range": f"bytes={mid}-{mid}"}
            response = httpx_client.get(url, headers=headers, follow_redirects=True)
            if response.status_code == 206:
                lower_bound = mid
                if upper_bound == lower_bound + 1:
                    return upper_bound
                if "Content-Range" in response.headers:
                    content_range = response.headers["Content-Range"]
                    total_size = int(content_range.split("/")[-1])
                    return total_size
            elif response.status_code == 416:
                upper_bound = mid
            else:
                raise Exception(
                    f"Unexpected status code: {response.status_code}",
                )
            if upper_bound - lower_bound <= 1:
                return upper_bound
            if upper_bound == initial_guess:
                upper_bound *= 2
        raise Exception("Cannot estimate remote resource size")


def get_resource_size(
    url: str,
    user_agent: Optional[str] = None,
) -> int:
    """
    Get the size of a remote resource.

    This function attempts to determine the size of a remote resource using different methods.
    It first tries to get the content length using a basic request. If that fails, it checks
    if the server accepts byte range requests. If so, it estimates the content length dynamically.

    Args:
        url (str): The URL of the remote resource.

    Returns:
        int: The size of the remote resource in bytes.

    Raises:
        Exception: If the size of the remote resource cannot be determined.
    """
    # headers = dict()
    # headers = _add_custom_user_agent_to_headers(user_agent, headers)
    try:
        size = request_content_length(url, user_agent=user_agent)
        return size
    except:
        range_accepted = check_if_remote_accepts_byte_range(url, user_agent=user_agent)
        if range_accepted:
            size = estimate_resource_size(url, user_agent=user_agent)
            return size
        else:
            raise Exception("cannot estimate remote resource size")


def save_file(
    url: str,
    path: str | Path,
    overwrite: bool = False,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool = True,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> None:
    """
    Download a remote resource to a local file.

    This function downloads a resource from a given URL and saves it to a specified local path.
    It supports progress tracking, overwrite protection, and redirect following.

    Args:
        url (str): The URL of the remote resource to download.
        path (str | Path): The local path where the downloaded resource will be saved.
        allow_overwrite (bool, optional): If True, allows overwriting existing files. Defaults to False.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        show_progress (bool, optional): If True, displays a progress bar during download. Defaults to False.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.

    Raises:
        FileExistsError: If the target file already exists and allow_overwrite is False.
        Exception: If there's an error during the download process.

    Returns:
        None
    """
    path = Path(path)
    if path.exists() and not overwrite:
        raise FileExistsError(str(path))
    size = None
    if evaluate_size:
        size = get_resource_size(url)
    file_name = utils.extract_file_name_from_url(url)
    with open(path, "wb") as file:
        with tqdm(
            total=size,
            unit_scale=True,
            unit="B",
            unit_divisor=chunk_size,
            disable=not show_progress,
            desc=file_name if show_file_name else None,
            ncols=80,
        ) as bar:
            with httpx.stream(
                "GET", url, follow_redirects=follow_redirects
            ) as response:
                for chunk in response.iter_bytes(chunk_size):
                    file.write(chunk)
                    bar.update(len(chunk))
    file.close()


def load_binary(
    url: str,
    chunk_size: int = 1 << 10,
    follow_redirects: bool = True,
    evaluate_size: bool = True,
    show_progress: bool = False,
    show_file_name: bool = True,
) -> bytes:
    """
    Read a remote resource and return its content as bytes.

    This function downloads a resource from a given URL and returns its content as bytes.
    It supports progress tracking and redirect following.

    Args:
        url (str): The URL of the remote resource to read.
        chunk_size (int, optional): The size of chunks to use when downloading. Defaults to 1024 bytes.
        show_progress (bool, optional): If True, displays a progress bar during download. Defaults to False.
        follow_redirects (bool, optional): If True, follows HTTP redirects. Defaults to True.

    Returns:
        bytes: The content of the remote resource.

    Raises:
        Exception: If there's an error during the download process.
    """
    size = None
    if evaluate_size:
        size = get_resource_size(url)
    buffer = BytesIO()
    file_name = utils.extract_file_name_from_url(url)
    with tqdm(
        total=size,
        unit_scale=True,
        unit="B",
        unit_divisor=chunk_size,
        disable=not show_progress,
        desc=file_name if show_file_name else None,
        ncols=80,
    ) as bar:
        with httpx.stream("GET", url, follow_redirects=follow_redirects) as response:
            for chunk in response.iter_bytes(chunk_size):
                buffer.write(chunk)
                bar.update(len(chunk))
    buffer.seek(0)
    data = buffer.read()
    buffer.close()
    return data


# def get(
#     url: str,
#     show_progress: bool = False,
#     user_agent: Optional[str] = None,
# ) -> bytes:
#     """_summary_

#     Args:
#         url (str): _description_
#         progress (bool, optional): _description_. Defaults to False.

#     Returns:
#         bytes: _description_
#     """
#     headers = dict()
#     headers = _add_custom_user_agent_to_headers(user_agent, headers)
#     raise NotImplementedError()


# def download(
#     url: str,
#     path: str | Path,
#     show_progress: bool = False,
#     user_agent: Optional[str] = None,
# ) -> None:
#     """_summary_

#     Args:
#         url (str): _description_
#         path (str | Path): _description_
#         progress (bool, optional): _description_. Defaults to False.
#     """
#     headers = dict()
#     headers = _add_custom_user_agent_to_headers(user_agent, headers)
#     raise NotImplementedError()



def parse(): ...


def serialize(): ...


def open(): ...


def read(): ...


def save(): ...


def load(): ...


def stream(): ...


def download(): ...


# def append(): ...


# def prepend(): ...


# def walk(): ...


# import orjson
# from pyeio.types import FilePath


# def stream(): ...


# # import json
# # from pathlib import Path
# # from .core import io


# # def load(path: str | Path) -> list | dict:
# #     return json.loads(io.load_text(path))


# # def save(): ...


# # todo
# # def stream(
# #     path: str | Path,
# #     validator,
# #     handler,
# # ): ...


# # # same as stream except operation is done on each element
# # def stream_ingest(
# #     path,
# #     validator,
# #     handler,
# #     errors,  # raise or log
# # ): ...



from importlib.metadata import version

__name__ = __package__ = __module__ = "pyeio"
__version__ = version(__name__)
__repository__ = "https://github.com/harttraveller/pyeio"
__documentation__ = "https://harttraveller.github.io/pyeio/"
__issues__ = "https://github.com/harttraveller/pyeio/issues/"



import re
from pathlib import Path
from typing import Pattern
from pyeio.common.types import FilePath


def open_text(path: FilePath) -> str:
    """
    Open a text file as a string.

    Args:
        path (FilePath): Path to the file.

    Returns:
        str: The data in the file.

    Raises:
        FileNotFoundError: Raised when file is not found.
    """
    with open(path, "r") as file:
        data = file.read()
    file.close()
    return data


def open_text_chars(path: FilePath) -> list[str]:
    """
    Open a text file and read its characters into a list of single characters.

    Args:
        path (FilePath): Path to the file.

    Returns:
        list[str]:  A list containing each character from the file as a string.

    Raises:
        FileNotFoundError: Raised when the file is not found at the specified path.
    """
    data = open_text(path=path)
    chars = list(data)
    return chars


# todo: adjust to stream in data and only keep relevant data in memory to reduce memory usage for large files
# todo: also add progress to all of the `open` functions


# todo.features
# drop_empty: drop empty strings
# logical (if multiple patterns, need to account for combinations of patterns, all, any, not, or, and, etc)
# keep_only: keep only groups containing matching patterns
# drop_match: drop groups containing matching patterns
def open_text_split(
    path: FilePath,
    delimiter: str,
    keep_delimiter: bool = True,
) -> list[str]:
    """
    Open a text file and split it into groups on a regex delimiter.

    Args:
        path (FilePath): Path to the file.
        delimiter (str): The regex string split the file into groups.
        keep (bool, optional): Whether to keep the delimiter. Defaults to True.

    Returns:
        list[str]: _description_
    """
    ...


# def open_text_tokens(
#     path: FilePath,
#     match: str | None,
#     groups
# )


def open_text_lines(path: FilePath) -> list[str]:
    """
    Open a text file and read its lines into a list of strings.

    Args:
        path (FilePath): Path to the file.

    Returns:
        list[str]: A list containing each line from the file as a string.

    Raises:
        FileNotFoundError: Raised when the file is not found at the specified path.
    """
    data = open_text(path=path)
    lines = data.split("\n")
    return lines


def save_text(
    data: str,
    path: str | Path,
    overwrite: bool = False,
) -> None:
    path = Path(path)
    if path.exists() and (not overwrite):
        raise FileExistsError(str(path))
    with open(path, "w") as file:
        file.write(data)
    file.close()


# def find_text(): ...


# def save_text_lines(): ...


# def read_text(): ...


# def read_text_lines(): ...


# def read_text_chars(): ...


# def read_text_chunk(path: FilePath, start: int, end: int) -> str: ...


# def read_text_segments():
#     """set start and end delimiters and stream the text between these"""
#     ...


# def append_text(): ...


# def prepend_text(): ...


# def insert_text(): ...


# def delete_text(): ...


# def append_text_line(): ...


# def append_text_lines(): ...


# def prepend_text_line(): ...


# def prepend_text_lines(): ...


# def insert_text_line(): ...


# def insert_text_lines(): ...


# def delete_text_line(): ...


# def delete_text_lines(): ...


# # def load_binary(): ...


# def open_binary_chunk(path: FilePath, start: int = 0, end: int | None = None) -> bytes:
#     """
#     start byte index
#     end byte index
#     """
#     ...


# def save_binary(): ...

# def open_text_chunk(path: FilePath, start: int = 0, end: int | None = None):
#     """
#     start char index
#     end char index
#     """
#     ...

# todo: add append and prepend data functions
# todo: add binary functions
# todo: add analogous web functions (load, stream)



import re

# todo: need to implement tokenizer that can handle generator for dynamic tokenization


class Tokenizer:
    def __init__(self, patterns, require_match: bool):
        # patterns should be a list of tuples in the format (token_type, pattern)
        self.patterns = [
            (token_type, re.compile(pattern)) for token_type, pattern in patterns
        ]
        self.require_match = require_match

    def tokenize(self, text):
        tokens = []
        pos = 0

        while pos < len(text):
            match = None
            for token_type, pattern in self.patterns:
                match = pattern.match(text, pos)
                if match:
                    token_text = match.group(0)
                    tokens.append((token_type, token_text))
                    pos = match.end()  # Move past the matched token
                    break
            if (not match) and (self.require_match):
                raise ValueError(f"Unexpected character at position {pos}: {text[pos]}")
            else:
                pos += 1

        return tokens



# uses result, returns a higher level object that operations can be applied to
# eg: CSV class, with append, prepend methods, type validation for data, automatically prepends after first column if column defined, etc


import pandas as pd
from pathlib import Path


def open(path: str | Path) -> pd.DataFrame:
    return pd.read_csv(path)


def save(data: pd.DataFrame, path: str | Path, overwrite: bool = False): ...


def get(url: str) -> pd.DataFrame: ...


# ! ---


class Streamer:
    def __init__(self, url: str) -> None:
        self.url = url


class Reader:
    def __init__(self, file: str | Path) -> None:
        self.file = Path(file)
        self.__column_names: list[str] | None = None

    @property
    def column_names(self) -> list[str]: ...

    def read_rows(self): ...

    def read_column_names(self): ...



import importlib
from pathlib import Path
from typing import Any, Literal
from pyeio.common.types import FileFormat

# todo.fix: figure out how to do proper type hints
# todo: need to add format identification and dynamic module loading based on format to avoid taking a massive amount of time to import
# todo: should


def parse(
    data: str | bytes, format: FileFormat | Literal["detect"] = "detect"
) -> Any: ...


def serialize(data, format: FileFormat = "json") -> str: ...


def open(path: str | Path) -> Any: ...


def save(): ...


def load(): ...


def download(): ...


def decompress(): ...


def compress(): ...


# todo.feature:
# * add a function named (connect/interface/?)? which allowes user to open dynamic format type from base schema File, with expanded functionality
#   * need to account for on disk file, or URL



from pathlib import Path
from typing import cast
from urllib.parse import urlparse
from pyeio.common import data
from pyeio.common.types import (
    FilePath,
    FileFormat,
    FileExtension,
    StandardFileExtension,
)
from pyeio.common.exceptions import (
    UnsupportedFileExtensionError,
    IncorrectFileExtensionError,
    MissingFileExtensionError,
)


def extract_raw_file_extension_from_path(path: FilePath) -> str | None:
    if isinstance(path, str):
        path = Path(path)
    file_name_components: list[str] = path.name.split(".")
    if len(file_name_components) == 1:
        return None
    else:
        file_extension = file_name_components[-1]
        return file_extension


def extract_raw_file_extension_from_url(url: str) -> str | None:
    parsed_url = urlparse(url)
    url_path = parsed_url.path
    file_extension = extract_raw_file_extension_from_path(path=url_path)
    return file_extension


def extract_file_name_from_path(path: FilePath) -> str:
    return Path(path).name


def extract_file_name_from_url(url: str) -> str:
    return urlparse(url).path.split("/")[-1]


def run_file_extension_validation(
    module_name: FileFormat,
    url: str | None = None,
    path: FilePath | None = None,
) -> None:
    if url:
        file_extension = extract_raw_file_extension_from_url(url)
        file_name = extract_file_name_from_url(url)
    elif path:
        file_extension = extract_raw_file_extension_from_path(path)
        file_name = extract_file_name_from_path(path=path)
    else:
        raise ValueError("Either `url` or `path` parameter must be passed.")
    if file_extension is None:
        raise MissingFileExtensionError(file=file_name)
    file_extension = file_extension.lower()
    if file_extension not in data.file_extensions:
        raise UnsupportedFileExtensionError(extension=file_extension)
    file_extension = cast(FileExtension, file_extension)
    allowed_variant_extensions = data.allowed_variants_lookup[module_name]
    if not (file_extension) in allowed_variant_extensions:
        raise IncorrectFileExtensionError(
            file_extension=file_extension,
            compatible_extensions=set(map(str, allowed_variant_extensions)),
        )


# def correct_file_extension(
#     real_extension: str | FileExtension,
#     expected_extension: FileExtension,
# ) -> None:
#     real_extension = real_extension.lower()
#     allowed_extensions: set[FileExtension] = index.allowed_variants_lookup[
#         real_extension
#     ]


# # todo: this should be refactored into a extension checking engine that includes warnings
# # for technically allowed file extensions, the ability to bypass checks with parameters, etc
# def is_valid_file_extension(extension: str, allowed: set[str]) -> bool:
#     return extension.lower() in allowed

# def raise_invalid_file_extension(extension: str, allowed: set[str]) -> None:


# todo
def rename_file():
    # todo: rename file on disk
    ...


def get_file_metadata():
    # todo: get builtin file metadata
    ...



from pytest import raises
from pyeio import json
from pyeio.common.exceptions import (
    MissingFileExtensionError,
    UnsupportedFileExtensionError,
    IncorrectFileExtensionError,
)


def test_parse_types():
    with raises(TypeError):
        _ = json.parse(data={"asdf", "1234"})  # type: ignore



from typing import get_args
from pyeio.common.types import (
    StandardFileExtension,
    VariantFileExtension,
    FileExtension,
)

standard_file_extensions: tuple[StandardFileExtension, ...] = get_args(
    StandardFileExtension
)
variant_file_extensions: tuple[VariantFileExtension, ...] = get_args(
    VariantFileExtension
)
file_extensions: tuple[FileExtension, ...] = (
    standard_file_extensions + variant_file_extensions
)


variant_to_standard: dict[VariantFileExtension, StandardFileExtension] = {
    "ndjson": "jsonl",
    "jsonlines": "jsonl",
    # "jpg": "jpeg",
    # "yml": "yaml",
    # "markdown": "md",
}


# used file extension : set { valid file extensions }
allowed_variants_lookup: dict[StandardFileExtension, set[FileExtension]] = {
    "json": {"json"},
    "jsonl": {
        "jsonl",
        "json",
        "jsonlines",
        "ndjson",
    },
}
